#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <gnutls/gnutls.h>

#define MAX_BUF 1024
#define CAFILE "server.req"
#define MSG "Tajna zprava\r\n"

  //certifikat CA
  gnutls_certificate_credentials xcred;


/* 
  Vytvoreni socketu a pripojení k serveru
 */
int tcp_connect(char *ip)
{
   int err, sd;
   struct sockaddr_in sa;

   //vytvoreni socketu
   sd = socket(AF_INET, SOCK_STREAM, 0);

   memset(&sa, '\0', sizeof(sa));
   sa.sin_family = AF_INET;
   sa.sin_port = htons(23987);

   //preklad ip adresy   
   struct hostent *phe;
   if((phe = gethostbyname(ip)) != NULL)
     memcpy((char *)&sa.sin_addr, phe->h_addr, phe->h_length);
   else if((sa.sin_addr.s_addr = inet_addr(ip)) == INADDR_NONE) {
     printf("Nelze ziskat ip adresu ani jmeno serveru\n");
     exit(1);
   }

   //inet_pton(AF_INET, ip, &sa.sin_addr);

   //pripojení k serveru
   err = connect(sd, (struct sockaddr *) & sa, sizeof(sa));
   if (err < 0) {
      printf("Nelze se pripojit\n");
      exit(1);
   }

   return sd;
} //tcp_connect

/* 
  Uzavre TCP spojeni
 */
void tcp_close( int sd) 
{
  shutdown(sd, SHUT_RDWR);
  close(sd);
}

/*
  Uzavre TLS spojeni
*/
void close_tls_session(int sd, gnutls_session session) {
   
   tcp_close(sd);

   //uvolni TLS spojeni
   gnutls_deinit(session);

   //uvolni certifikat
   gnutls_certificate_free_credentials(xcred);

   //volano kvuli ostatnim knihovnam (napr. libgrypt)
   gnutls_global_deinit();
}

/*
  Jednoduchý TLS klient s pouzitim GnuTLS.
  Posle zpravu serveru, ten ji prijme a posle nazpet.
  Jako parametr se zadá IP adresa serveru.
*/
int main(int argc, char **args)
{
   int ret, sd;
   gnutls_session session;
   char buffer[MAX_BUF + 1];
   
   //Chci certifikat X.509, pripadne OpenPGP, v nejhorsim pripade bez autentizace
   const int cert_type_priority[3] = { GNUTLS_CRT_X509, 
                                       GNUTLS_CRT_OPENPGP,
                                       0
                                     };
//   const int cert_type_priority[1] = { 0 };

   if(argc != 2) {
     printf("Program se spousti s jedinym parametrem.\nZadejte jmeno serveru nebo IP adresu.\n");
     return 5;
   }   
 
   //inicializace promennych z libgrypt
   gnutls_global_init();

   // X509 certifikat
   gnutls_certificate_allocate_credentials(&xcred);

   //nastaveni cetifikacni autority, kterou akceptuje
   gnutls_certificate_set_x509_trust_file(xcred, CAFILE, GNUTLS_X509_FMT_PEM);

   // inicializace TLS spojeni 
   gnutls_init(&session, GNUTLS_CLIENT);

   //nastavi vychozi prirority spojeni
   gnutls_set_default_priority(session);
   gnutls_certificate_type_set_priority(session, cert_type_priority);

   //priradi certifikat
   gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, xcred);

   //pripojeni      
   sd = tcp_connect(args[1]);

   //TLS spojeni session bude využívat TCP spojení dané číslem sd
   gnutls_transport_set_ptr(session, (gnutls_transport_ptr)sd);

   //provede handshake
   ret = gnutls_handshake(session);
   if(ret < 0) {
     printf("Nastala chyba behem handshaku\n");
     gnutls_perror(ret);
     close_tls_session(sd, session);
     return 1;
   }
   else
     printf("Handshake probehl uspesne.\n");

   //zaslání zprávy pres TLS
   printf("Posilam zpravu: %s\n", MSG);
   gnutls_record_send(session, MSG, strlen(MSG));

   //příjem zprávy
   ret = gnutls_record_recv( session, buffer, MAX_BUF);
   if(ret == 0) {
     printf("Server necekane ukoncil TLS spojeni.\n");
     close_tls_session(sd, session);
     return 2;
   }
   else if (ret < 0) {
     printf("Chyba: %s\n", gnutls_strerror(ret));
     close_tls_session(sd, session);
     return 3;
   }
   printf("Prijato %d bytu: ", ret);
   buffer[ret] = '\0';
   printf("%s \n", buffer);

   //dame vedet serveru ze koncime   
   gnutls_bye(session, GNUTLS_SHUT_RDWR);

   //konec spojeni
   close_tls_session(sd, session);
   return 0;
} //main

